#!/usr/bin/env python
import sys

if len(sys.argv) > 1 and sys.argv[1] == '--version':
    print('Version 0.1.1', file=sys.stderr)
    sys.exit(0)

if len(sys.argv) != 2 + 1:
    sys.exit(1)

print(f'a+b = {sys.argv[1]} + {sys.argv[2]}')



""" String filters for Baygon.
Each filter is a class that implements a filter method.
A filter is used to modify `stdout` and `stderr` before they are tested.
"""
import re
from abc import ABC, abstractmethod
from collections.abc import Sequence
from .error import InvalidFilterError


class Filter(ABC):
    """ Base class for filters. """
    @abstractmethod
    def filter(self, value: str) -> str:
        """ Return True if the value matches the filter. """
        return value

    def __init__(self, name: str = None):
        self.name = name

    def __repr__(self):
        return f'{self.__class__.__name__}<{self.name}>'

    def __call__(self, value: str) -> str:
        return self.filter(value)


class FilterNone(Filter):
    """ Filter that does nothing. """

    def filter(self, value: str) -> str:
        return value


class FilterUppercase(Filter):
    """ Filter for uppercase strings.
    >>> f = FilterUppercase()
    >>> f('hello')
    'HELLO'
    """
    type = 'uppercase'

    def filter(self, value: str) -> str:
        return value.upper()


class FilterLowercase(Filter):
    """ Filter for lowercase strings.
    >>> f = FilterLowercase()
    >>> f('HELLO')
    'hello'
    """

    type = 'lowercase'

    def filter(self, value: str) -> str:
        return value.lower()


class FilterTrim(Filter):
    """ Filter for trimmed strings.
    >>> f = FilterTrim()
    >>> f(' hello   ')
    'hello'
    """
    type = 'trim'

    def filter(self, value: str) -> str:
        return value.strip()


class FilterIgnoreSpaces(Filter):
    """ Filter for strings with no spaces.
    >>> f = FilterIgnoreSpaces()
    >>> f('hello   world')
    'helloworld'
    """
    type = 'ignorespaces'

    def filter(self, value: str) -> str:
        return value.replace(' ', '')


class FilterReplace(Filter):
    """ Filter for strings with simple replacements.
    >>> f = FilterReplace('hello', 'world')
    >>> f('hello world')
    'world world'
    """
    type = 'replace'

    def __init__(self, pattern: str, replacement: str):
        super().__init__()
        self.pattern = pattern
        self.replacement = replacement

    def filter(self, value: str) -> str:
        return value.replace(self.pattern, self.replacement)


class FilterRegex(Filter):
    """ Filter for strings using regular expressions.
    >>> f = FilterRegex('[aeiou]', '-')
    >>> f('hello world')
    'h-ll- w-rld'
    """
    type = 'regex'

    def __init__(self, pattern: str, replacement: str):
        super().__init__()
        self.pattern = pattern
        self.replacement = replacement
        self.regex = re.compile(pattern)

    def filter(self, value: str) -> str:
        return self.regex.sub(self.replacement, value)


class Filters(Filter, Sequence):
    """ A sequence of filters. """
    type = 'filters'

    def __init__(self, filters=None):
        super().__init__()
        self._filters = self._parse_filter(filters)

    def _parse_filter(self, filters):

        if filters is None:
            return []
        if isinstance(filters, Filter):
            return [filters]
        if isinstance(filters, Filters):
            return filters._filters
        if isinstance(filters, dict):
            return [filter_map[name](args)
                    for name, args in filters.items()]

        raise InvalidFilterError(f'Invalid type for filters: {type(filters)}')

    def __getitem__(self, index):
        return self._filters[index]

    def __len__(self):
        return len(self._filters)

    def extend(self, filters):
        """ Extend the filters with another Filters object. """
        self._filters.extend(self._parse_filter(filters))
        return self

    def filter(self, value: str) -> str:
        for filter_ in self._filters:
            value = filter_.filter(value)
        return value

    def __repr__(self):
        return f'{self.__class__.__name__}<{self._filters}>'


filter_map = {
    'uppercase': FilterUppercase,
    'lowercase': FilterLowercase,
    'trim': FilterTrim,
    'ignorespaces': FilterIgnoreSpaces,
    'replace': FilterReplace,
    'regex': FilterRegex,
}



""" Test config file. """
from pathlib import Path
from unittest import TestCase

from baygon.suite import TestSuite

dir_path = Path(__file__).resolve(strict=True).parent


class TestConfig(TestCase):
    def test_config_file(self):
        ts = TestSuite(path='tests')
        self.assertIn('t.yml', str(ts.path))
        self.assertIn('version', ts.config)
        self.assertEqual(ts.config['version'], 1)



""" YAML Schema for Baygon test files. """
from voluptuous import All, Any, Boolean, Coerce, ExactSequence, Optional, Required
from voluptuous import Schema as VSchema
from voluptuous import Self
from voluptuous.humanize import validate_with_humanized_errors

from .id import Id


class TrackId:
    """ Keep the id of the test. """

    def __init__(self):
        self._id = Id()

    def reset(self):
        """ Reset the id. """
        def _(v):
            self._id = Id()
            return v
        return _

    def down(self):
        """ Return a new Id with the given id appended. """
        def _(v):
            self._id = self._id.down()
            return v
        return _

    def up(self):
        """ Return a new Id with the given id appended. """
        def _(v):
            self._id = self._id.up()
            return v
        return _

    def next(self):
        """ Return a new Id with the last id incremented. """
        def _(v):
            v['test_id'] = list(self._id)
            self._id = self._id.next()
            return v
        return _


Value = Any(str,
            All(Any(int, float, All(bool, Coerce(int))), Coerce(str)))

# Global test filters
filters = {
    Optional('uppercase'): Boolean(),
    Optional('lowercase'): Boolean(),
    Optional('trim'): Boolean(),
    Optional(Any('ignorespaces', 'ignore-spaces')): Boolean(),
    Optional('regex'): ExactSequence([str, str]),
    Optional('replace'): ExactSequence([str, str]),
}

# Test case
case = {
    Optional('filters', default={}): filters,
    Any('equals', 'regex', 'contains'): Value,
    Optional('not'): [{Required(Any('equals', 'regex', 'contains')): Value}],
    Optional('expected',
             description="Expected value when used with regex"): Value,
}

# Nested test cases
match = Any(
    All(Value, lambda x: [{'equals': x}]),
    [
        All(Value, lambda x: {'equals': x}),
        case
    ],
    case,
)

common = {
    Optional('name',
             default='',
             description='Test Name'): str,
    Optional('executable',
             default=None,
             description='Path to the executable'): Any(str, None),
    Optional('points',
             default=0,
             description="Points given for this test"): int,
}

test = VSchema({
    Optional('args', default=[]): [Value],
    Optional('env', default={}): {str: Value},

    Optional('stdin', default=''): Any(Value, None),
    Optional('stdout', default=[]): match,
    Optional('stderr', default=[]): match,

    Optional('exit'): All(Any(int, Boolean()), Coerce(int))
}).extend(common)

Num = TrackId()

group = VSchema({
    Required('tests'): All(Num.down(),
                           [All(Any(Self, test), Num.next())],
                           Num.up())
}).extend(common)


def Schema(data, humanize=False):  # noqa: N802
    """ Validate the given data against the schema. """
    schema = VSchema({
        Optional('name'): str,
        Optional('version', default=2): Any(1, 2),
        Optional('filters', default={}): filters,
        Required('tests'): All(Num.reset(),
                               [All(Any(test, group), Num.next())])
    }).extend(common)
    if humanize:
        return validate_with_humanized_errors(data, schema)
    return schema(data)



""" Test suite. """
import json
from pathlib import Path
from .error import InvalidExecutableError, ConfigError
import yaml

from . import error
from .executable import Executable
from .filters import Filters
from .id import Id
from .schema import Schema
from .str import GreppableString


def find_testfile(path=None):
    """Recursively find the tests description file."""
    if not path:
        path = Path('.')
    elif isinstance(path, str):
        path = Path(path)

    path = path.resolve(strict=True)

    if path.is_file():
        return path

    for filename in ['baygon', 't', 'test', 'tests']:
        for ext in ['json', 'yml', 'yaml']:
            f = path.joinpath(f"{filename}.{ext}")
            if f.exists():
                return f

    # Recursively search in parent directories
    if path.parent == path:  # Test if root directory
        return None

    return find_testfile(path.parent)


def load_config(path=None):
    """Load a configuration file (can be YAML or JSON)."""
    path = find_testfile(path)

    if not path.exists():
        raise ConfigError(
            f"Couldn't find and configuration file in '{path.resolve()}'")

    with open(path, 'rt', encoding="utf-8") as fp:
        if path.suffix in ['.yml', '.yaml']:
            return Schema(yaml.safe_load(fp))
        if path.suffix in ['.json']:
            return Schema(json.load(fp))

    raise ConfigError(f"Unknown file extension '{path.suffix}' for '{path}'")


class BaseMixin:
    """ Base mixin to prevent super() failure.
    Ensure it will be the last MRO (Method Resolution Order)."""

    def __init__(self, *args, **kwargs):
        pass


class FilterMixin(BaseMixin):
    """ Mixin for tests that can have filters. """

    def __init__(self, *args, **kwargs):
        config = args[0]

        if 'parent' in kwargs and hasattr(kwargs['parent'], 'filters'):
            self.filters = kwargs['parent'].filters
        else:
            self.filters = Filters()

        if 'filters' in config:
            self.filters.extend(config['filters'])

        super().__init__(*args, **kwargs)


class ExecutableMixin(BaseMixin):
    """ Mixin for tests that have an executable. """

    def __init__(self, *args, **kwargs):
        config = args[0]
        self.cwd = kwargs['parent'].cwd

        if hasattr(kwargs['parent'], 'executable'):
            executable = kwargs['parent'].executable
        else:
            executable = None

        if executable is not None and config['executable'] is not None:
            raise InvalidExecutableError(
                "Executable can't be overridden")

        if config['executable'] is not None:
            exe = self.cwd.joinpath(config['executable']).resolve(strict=True)
            self.executable = Executable(exe)
        else:
            self.executable = Executable(executable)

        super().__init__(*args, **kwargs)


class NamedMixin(BaseMixin):
    """ Mixin for tests having a name. """

    def __init__(self, *args, **kwargs):
        config = args[0]

        self.name = config['name']
        self.id = Id(config['test_id'])
        self.points = config['points']

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.id}. {self.name}>"


class GroupMixin(BaseMixin):
    """ Mixin for tests that can have sub-tests. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = args[0]
        self.tests = []
        for test in config['tests']:
            if 'tests' in test:
                self.tests.append(TestGroup(test, parent=self))
            else:
                self.tests.append(TestCase(test, parent=self))

    def __len__(self):
        return len(self.tests)

    def __getitem__(self, item):
        return self.tests[item]

    def run(self, flatten=False):
        """ Run the tests. """
        if flatten:
            issues = []
            for test in self.tests:
                run = test.run()
                issues += (run if isinstance(run, list) else [run])
        else:
            issues = [test.run() for test in self.tests]
        return issues

    def get_points(self):
        """ Return the total number of points. """
        return sum(test.points for test in self.tests)


class TestCase(NamedMixin, ExecutableMixin, FilterMixin):
    """ A single test. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        config = args[0]

        self.args = config.get('args', [])
        self.env = config.get('env', {})

        self.stdin = config.get('stdin', '')
        self.stdout = config.get('stdout', [])
        self.stderr = config.get('stderr', [])

        self.exit = config.get('exit', None)
        self.config = config
        self.output = None
        self.issues = []

    def run(self):
        """ Run the tests. """
        if not isinstance(self.executable, Executable):
            raise InvalidExecutableError(
                f"Test {self.id}, not a valid executable: {self.executable}")

        self.output = output = self.executable.run(*self.args,
                                                   stdin=self.stdin)

        self.issues = [
            *self._check_exit_status(output.exit_status),
            *self._check_stdout(output.stdout),
            *self._check_stderr(output.stderr)
        ]
        return self.issues

    def _check_exit_status(self, output):
        if self.exit is None:  # No exit status specified
            return []
        if output != self.exit:
            return [error.InvalidExitStatus(output, self.exit,
                                            name=self.name, id=self.id)]
        return []

    def _check_stdout(self, output):
        return self._match(self.stdout, output, where='stdout')

    def _check_stderr(self, output):
        if self.stderr is None:
            return []
        return self._match(self.stderr, output, where='stderr')

    def _match(self, options: list, value: str,
               where=None, inverse=False) -> list:
        """ Match a value against a list of options. """
        issues = []
        for case in options:
            value = GreppableString(self.filters.extend(case.get('filters', {}))(value))

            if 'regex' in case:
                if (not value.grep(case['regex'])) ^ inverse:
                    issues += [error.InvalidRegex(value, case['regex'],
                                                  on=where, name=self.name,
                                                  id=self.id)]
            if 'contains' in case:
                if (not value.contains(case['contains'])) ^ inverse:
                    issues += [
                        error.InvalidContains(
                            value, case['contains'], on=where,
                            name=self.name, id=self.id)]
            if 'equals' in case:
                if (case['equals'] != value) ^ inverse:
                    issues += [
                        error.InvalidEquals(value, case['equals'], on=where,
                                            name=self.name, id=self.id)]
            if 'not' in case:
                issues += self._match(case['not'], value, where, inverse=True)

        return issues


class TestGroup(NamedMixin, ExecutableMixin, FilterMixin, GroupMixin):
    """ A group of tests. """


class TestSuite(ExecutableMixin, FilterMixin, GroupMixin):
    """ Test suite. """
    __test__ = False  # Don't run this class as a test

    def __init__(self, data: dict = None, path=None, executable=None, cwd=None):
        if isinstance(data, dict):
            self.config = Schema(data)
            cwd = Path.cwd()
        else:
            self.path = find_testfile(path)
            cwd = self.path
            self.config = load_config(self.path)

        self.name = self.config.get('name', 'Test Suite')
        self.version = self.config.get('version')

        class Root:
            def __init__(self, executable):
                self.executable = Executable(executable)
                self.cwd = cwd.resolve(strict=True).parent

        super().__init__(self.config, parent=Root(executable))



""" Errors for Baygon """


class BaygonError(Exception):
    """ Base class for Baygon errors """
    pass


class ConfigError(BaygonError):
    """ Raised when a config value is not valid """
    pass


class InvalidExecutableError(BaygonError):
    """ Raised when an executable is not found """
    pass


class InvalidFilterError(BaygonError):
    """ Raised when a filter is not found """
    pass


class InvalidCondition:
    """ Invalid condition error. """

    def __init__(self, value, expected, message=None, on=None, **kwargs):
        self.value = value
        self.expected = expected
        self.message = message
        self.name = kwargs.get('name', '')
        self.id = kwargs.get('id', None)
        self.on = on

    def __str__(self):
        return f'Expected "{self.expected}", but got "{self.value}"'


class InvalidExitStatus(InvalidCondition):
    """ Invalid exit status error. """

    def __str__(self):
        if hasattr(self.value, '__len__') and len(self.value) > 20:
            return f'Invalid exit status. Expected {self.expected}'

        return (f'Invalid exit status. '
                f'Expected {self.expected}, but got {self.value}')


class InvalidContains(InvalidCondition):
    """ Invalid contains error. """


class InvalidRegex(InvalidCondition):
    """ Invalid regex error. """

    def __str__(self):
        return (f'Invalid value on {self.on}. '
                f'Expected to match regex /{self.expected}/')


class InvalidEquals(InvalidCondition):
    """ Invalid equals error. """

    def __str__(self):
        if hasattr(self.value, '__len__') and len(self.value) > 20:
            return (f'Invalid value on {self.on}. '
                    f'Expected exactly "{self.expected}"')

        return (f'Invalid value on {self.on}. '
                f'Expected exactly "{self.expected}", '
                f'but got "{self.value}"')



""" Executable class. To be used with the Test class. """

import os
import shutil
import subprocess
import typing
from collections import namedtuple
from pathlib import Path
from .error import InvalidExecutableError

Outputs = namedtuple('Outputs', ['exit_status', 'stdout', 'stderr'])

forbidden_binaries = ['rm', 'mv', 'dd', 'wget', 'mkfs']


def get_env(env: typing.Optional[str] = None) -> dict:
    """ Get the environment variables to be used for the subprocess. """
    return {**os.environ, **(env or {})}


class Executable:
    """ An executable program.

    Convenient execution and access to program outputs such as:

        - Exit status
        - Standard output
        - Standard error

    For example:

        >>> e = Executable('echo')
        Executable<echo>
        >>> e('-n', 'Hello World')
        Outputs(exit_status=0, stdout='Hello World', stderr='')
        >>> e('-n', 'Hello World').stdout
        'Hello World!'
    """

    def __new__(cls, filename):
        if isinstance(filename, cls):
            return filename

        return super().__new__(cls) if filename else None

    def __init__(self, filename, encoding='utf-8'):
        """ Create an executable object.

        :param filename: The path of the executable.
        :param encoding: The encoding to be used for the outputs, default is UTF-8.
        """
        if isinstance(filename, self.__class__):
            self.filename = filename.filename
            self.encoding = filename.encoding
        else:
            self.filename = filename
            self.encoding = encoding

        if not self._is_executable(self.filename):
            if '/' not in filename and shutil.which(filename) is not None:
                if filename in forbidden_binaries:
                    raise InvalidExecutableError(f"Program '{filename}' is forbidden!")
                filename = shutil.which(filename)
            else:
                raise InvalidExecutableError(f"Program '{filename}' is not an executable!")

    def run(self, *args, stdin=None, env=None):
        """ Run the program and grab all the outputs. """
        with subprocess.Popen([self.filename, *[str(a) for a in args]],
                              stdout=subprocess.PIPE,
                              stdin=subprocess.PIPE,
                              stderr=subprocess.PIPE, env=env) as proc:

            if stdin is not None:
                stdin = stdin.encode(self.encoding)

            stdout, stderr = proc.communicate(input=stdin)

            if stdout is not None:
                stdout = stdout.decode(self.encoding)
            if stderr is not None:
                stderr = stderr.decode(self.encoding)

            return Outputs(
                proc.returncode,
                stdout,
                stderr)

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.filename}>"

    @staticmethod
    def _is_executable(filename):
        path = Path(filename)
        return path.is_file() and os.access(path, os.X_OK)



import re


class GreppableString(str):
    """ A string that can be parsed with regular expressions. """

    def grep(self, pattern: str, *args) -> bool:
        """ Return True if the pattern is found in the string. """
        return re.findall(pattern, self, *args)

    def contains(self, value: str) -> bool:
        """ Return True if the value is found in the string. """
        return value in self



from unittest import TestCase

from baygon.suite import TestSuite


class TestDescription(TestCase):
    def get_sample(self, name='foobar'):
        return TestSuite({
            'name': name,
            'version': 1,
            'tests': [
                {'exit': 0}
            ]
        })

    def test_build(self):
        name = 'foobar'
        td = self.get_sample(name)
        self.assertEqual(td.name, name)
        self.assertEqual(td.version, 1)

    def test_len(self):
        name = 'foobar'
        td = self.get_sample(name)
        self.assertEqual(len(td), 1)




from unittest import TestCase

from baygon.id import Id


class TestId(TestCase):
    def test_id(self):
        i = Id('1.2.3.4')
        self.assertEqual(str(i), '1.2.3.4')
        self.assertEqual(tuple(i), (1, 2, 3, 4))
        self.assertEqual(list(i), [1, 2, 3, 4])



""" Hierarchical Id class to identify nested sequences """
import re
from collections.abc import Sequence
from typing import List


class Id(Sequence):
    """ Test identifier. Helper class to number tests.

    Example:

    >>> i = Id()
    >>> i = i.down()
    >>> i
    Id(1.1)
    >>> i += 1
    >>> i
    Id(1.2)
    >>> i += 2
    >>> i
    Id(1.4)
    >>> str(i)
    '1.4'
    >>> i.up()
    Id(1)
    >>> ((i + 1).down() + 1).down()
    Id(1.5.2.1)
    >>> tuple(Id().down().next().down().next().down().next())
    (1, 2, 2, 2)
    """

    def __init__(self, ids: List[int] = None):
        if isinstance(ids, int):
            ids = [ids]
        if isinstance(ids, Id):
            ids = ids.ids
        if isinstance(ids, str) and re.match(r'^\d+(\.\d+)*$', ids):
            ids = [int(i) for i in ids.split('.')]
        if ids is not None and not isinstance(ids, list):
            raise ValueError(f"Invalid type for Id: {type(ids)}")

        self.ids = ids or [1]

    def next(self):
        """ Return a new Id with the last id incremented. """
        return self + 1

    def down(self, base: int = 1):
        """ Return a new Id with the given id appended. """
        return Id(self.ids + [base])

    def up(self):
        """ Return a new Id with the given id appended. """
        return Id(self.ids[:-1])

    def __str__(self):
        return '.'.join([str(i) for i in self.ids])

    def __repr__(self):
        return f'Id({str(self)})'

    def __add__(self, other):
        return Id(self.ids[:-1] + [self.ids[-1] + other])

    def __list__(self):
        return self.ids

    def __iter__(self):
        for i in self.ids:
            yield i

    def __len__(self):
        return len(self.ids)

    def __hash__(self):
        return hash(tuple(self))

    def __getitem__(self, item):
        return self.ids[item]

    def pad(self, length='  '):
        """ Return id with initial padding. """
        return length * (len(self) - 1)



""" Main CLI command. """
import logging
import os
import sys
import time

import click

from . import TestCase, TestGroup, TestSuite, __copyright__, __version__
from .error import InvalidExecutableError


def display_pad(pad=0):
    """ Display a pad of spaces to align nested output. """
    click.secho('.' * pad if pad > 0 else '  ', nl=False, dim=pad == 0)


def test_name_length(test):
    """ Compute the length of test name. """
    return len(f'{test.id.pad()}Test {test.id}: {test.name}')


def display_test_name(test):
    """ Display the name of a test. """
    click.secho(f'{test.id.pad()}Test {test.id}: ', nl=False, bold=True)
    click.secho(f'{test.name}', nl=False, bold=False)


class OneLineExceptionFormatter(logging.Formatter):
    """ A formatter that displays the exception on a single line. """

    def format_exception(self, exc_info):
        """ Format an exception. """
        result = super().format_exception(exc_info)
        return repr(result)

    def format(self, record):
        """ Format a log record. """
        result = super().format(record)
        if record.exc_text:
            result = result.replace("\n", "")
        return result


class Runner:
    """ Test runner. """

    def __init__(self, executable=None, config=None, **kwargs):
        self.verbose = kwargs.get('verbose', 0)
        if self.verbose > 3:
            self._init_logger("DEBUG")

        self.executable = executable
        self.limit = kwargs.get('limit', -1)
        click.secho(config, fg='yellow')
        self.test_suite = TestSuite(path=config, executable=self.executable)

        self.align_column = 0

        self.failures = 0
        self.successes = 0
        self.skipped = 0

    def _init_logger(self, loglevel):
        handler = logging.StreamHandler()
        formatter = OneLineExceptionFormatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        root = logging.getLogger()
        root.setLevel(os.environ.get("LOGLEVEL", loglevel))
        root.addHandler(handler)

    def run(self):
        """ Run the tests. """
        start_time = time.time()
        self.align_column = self._max_length(self.test_suite) + 10

        self.failures = 0
        self.successes = 0
        self.skipped = 0

        self._traverse_group(self.test_suite)

        tests = self.failures + self.successes + self.skipped
        seconds = round(time.time() - start_time, 2)
        click.secho(f'\nRan {tests} tests in {seconds}s.', bold=True)

        if self.failures > 0:
            ratio = 100 - round(
                self.failures / (self.failures + self.successes) * 100, 2)
            click.secho((f'{self.failures} failed, {self.successes} '
                         f'passed ({ratio}%% ok).'), fg='yellow', bold=True)
            click.secho('\nfail.', fg='red', bold=True)
        else:
            click.secho('\nok.', fg='green')

        if self.skipped > 0:
            click.secho((
                f'{self.skipped} test(s) skipped, '
                'some executables may be missing.'), fg='yellow')

        return self.failures

    def _max_length(self, tests):
        length = 0
        for test in tests:
            length = max(test_name_length(test), length)
            if isinstance(test, TestGroup):
                length = max(self._max_length(test), length)
        return length

    def _traverse_group(self, tests):
        for test in tests:
            if self.limit > 0 and self.failures > self.limit:
                break

            if isinstance(test, TestGroup):
                display_test_name(test)
                click.echo('')
                self._traverse_group(test)
            elif isinstance(test, TestCase):
                display_test_name(test)
                issues = test.run()
                if issues is None:  # Skipped
                    click.secho(' SKIPPED', fg='yellow')
                    self.skipped += 1
                    continue
                display_pad(self.align_column - test_name_length(test))
                if not issues:
                    self.successes += 1
                    click.secho(' PASSED', fg='green')
                else:
                    self.failures += 1
                    click.secho(' FAILED', fg='red', bold=True)
                    for issue in issues:
                        click.secho('  ' * len(test.id) + '- ' + str(issue),
                                    fg='magenta')
        return self.failures


def version():
    """ Display the version. """
    print(f"Baygon version {__version__} {__copyright__}")


@click.command()
@click.argument('executable', required=False, type=click.Path(exists=True))
@click.option('--version', is_flag=True, help='Shows version')
@click.option('-v', '--verbose', count=True, help='Shows more details')
@click.option('-l', '--limit', type=int, default=-1, help='Limit errors to N')
@click.option('-d', '--debug', is_flag=True, default=0, help='Debug mode')
@click.option('-t', '--config',
              type=click.Path(exists=True),
              help='Choose config file (.yml or .json)')
def cli(debug, verbose=0, executable=None, config=None, **kwargs):
    """ Baygon functional test runner. """
    if kwargs.get('version'):
        version()
        sys.exit(0)

    if debug:
        failures = Runner(executable, config, verbose=verbose, **kwargs).run()
    else:
        try:
            failures = Runner(executable, config, verbose=verbose, **kwargs).run()
        except InvalidExecutableError as error:
            click.secho(f"\nError: {error}", fg='red', bold=True, err=True)
            sys.exit(1)

    click.echo('')
    return failures


if __name__ == '__main__':
    cli()




from pathlib import Path
from unittest import TestCase

import baygon
from baygon import error

dir_path = Path(__file__).resolve(strict=True).parent


class TestDemo(TestCase):
    def setUp(self):
        self.exe = baygon.Executable(dir_path.joinpath('main.py'))

    def test_minimal(self):
        print(dir_path)
        ts = baygon.TestSuite(path=dir_path, executable=self.exe)
        self.assertEqual(ts.run(), [[[], []], [], []])

    def test_error_exit_status(self):
        ts = baygon.TestSuite(
            {'tests': [{'exit': 12}]},
            executable=self.exe
        )

        self.assertIsInstance(ts.run()[0][0], error.InvalidExitStatus)

    def test_error_stdout(self):
        ts = baygon.TestSuite(
            {'tests': [{'args': [30, 40], 'stdout': 42}]},
            executable=self.exe
        )
        self.assertIsInstance(ts.run()[0][0], error.InvalidEquals)

    def test_error_stderr(self):
        ts = baygon.TestSuite(
            {'tests': [{'args': ['--version'], 'stderr': 42}]},
            executable=self.exe
        )
        self.assertIsInstance(ts.run()[0][0], error.InvalidEquals)

    def test_contains_stderr(self):
        ts = baygon.TestSuite(
            {'tests': [{'args': ['--version'], 'stderr': [{'contains': 'Version'}]}]},
            executable=self.exe
        )
        self.assertEqual(ts.run(), [[]])

    def test_uppercase_filter_with_contains_stderr(self):
        ts = baygon.TestSuite(
            {
                'filters': {
                    'uppercase': True
                },
                'tests': [
                    {
                        'args': ['--version'],
                        'stderr': [{'contains': 'VERSION'}]
                    }
                ]
            },
            executable=self.exe
        )
        t = ts.run()

        self.assertEqual(t, [[]])

    def test_lowercase_filter_with_contains_stderr(self):
        ts = baygon.TestSuite(
            {
                'filters': {
                    'lowercase': True
                },
                'tests': [
                    {
                        'args': ['--version'],
                        'stderr': [{'contains': 'version'}]
                    }
                ]
            },
            executable=self.exe
        )
        t = ts.run()
        self.assertEqual(t, [[]])



import shutil
from pathlib import Path
from unittest import TestCase

from baygon import Executable
from baygon.str import GreppableString
from baygon.error import InvalidExecutableError

dir_path = Path(__file__).resolve(strict=True).parent


class TestExecutable(TestCase):
    def test_not_a_file(self):
        self.assertRaises(InvalidExecutableError, Executable, 'not-a-file')

    def test_arg(self):
        e = Executable(shutil.which('echo'))
        test_string = 'Live as if you were to die tomorrow'
        output = e.run('-n', test_string)
        print(output)
        self.assertEqual(output.stdout, test_string)

    def test_stdout(self):
        e = Executable(dir_path.joinpath('dummy.py'))
        output = e.run()
        print(output)
        self.assertEqual(output.stdout, "an apple\n")

    def test_stderr(self):
        e = Executable(dir_path.joinpath('dummy.py'))
        output = e.run()
        print(output)
        self.assertEqual(output.stderr, "an orange\n")

    def test_stdin(self):
        e = Executable(shutil.which('cat'))
        test_string = 'Live as if you were to die tomorrow'
        output = e.run(stdin=test_string)
        print(output)
        self.assertEqual(output.stdout, test_string)

    def test_exit_status(self):
        e = Executable(dir_path.joinpath('dummy.py'))
        output = e.run()
        print(output)
        self.assertEqual(output.exit_status, 42)

    def test_args(self):
        e = Executable(dir_path.joinpath('args.py'))
        test_string = 'foobar'
        output = e.run(2, test_string)
        print(output)
        self.assertEqual(output.stdout, test_string + '\n')

    def test_grep(self):
        s = GreppableString('Live as if you were to die tomorrow')
        u = s.grep(r'\b[aeiouy]\w{2}\b')
        print(u)
        self.assertEqual(u, ['you'])

    def test_echo(self):
        e = Executable('echo')
        test_string = 'Live as if you were to die tomorrow'
        output = e.run(test_string)
        print(output)
        self.assertEqual(output.stdout, test_string + "\n")

    def test_printf(self):
        e = Executable('printf')
        test_string = 'Live as if you were to die tomorrow'
        output = e.run(test_string)
        print(output)
        self.assertEqual(output.stdout, test_string)

    def test_cat(self):
        e = Executable('cat')
        output = e.run(dir_path.joinpath('test.txt'))
        print(output)
        self.assertEqual(output.stdout, "Hello World\n")



""" Baygon is a tool to run tests on executables. """
from .executable import Executable
from .schema import Schema
from .suite import TestCase, TestGroup, TestSuite

__all__ = ['TestCase', 'TestGroup', 'TestSuite', 'Executable', 'Schema']

try:
    from . import version
    __version__ = version.version
except ImportError:
    __version__ = None

__year__ = 2022
__copyright__ = f'Copyright {__year__} HEIG-VD'



from unittest import TestCase

from voluptuous.error import MultipleInvalid

from baygon import Schema


class TestSchema(TestCase):
    def test_minimal(self):
        Schema({
            'version': 1,
            'tests': [
                {'exit': 0}
            ]
        })

    def test_wrong_version(self):
        self.assertRaises(MultipleInvalid, Schema, {
            'version': 3,
            'tests': [
                {'exit': 0}
            ]
        })

    def test_filters(self):
        Schema({
            'version': 1,
            'filters': {
                'uppercase': True,
                'ignorespaces': True
            },
            'tests': [
            ]
        })

    def test_test_contains(self):
        Schema({'tests': [{'args': ['--version'], 'stderr': [{'contains': 'Version'}]}]})

    def test_empty_filters(self):
        s = Schema({
            'version': 1,
            'tests': []
        })
        self.assertIn('filters', s)



""" Test CLI. """
from pathlib import Path
from unittest import TestCase

from baygon.__main__ import cli
from click.testing import CliRunner


class TestVersion(TestCase):
    def test_version(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['--version'])

        self.assertEqual(result.exit_code, 0)
        self.assertIn('version', result.output)


class TestDemo(TestCase):
    @property
    def directory(self):
        return Path(__file__).resolve(strict=True).parent

    @property
    def executable(self):
        return str(self.directory.joinpath('main.py'))

    def get_config(self, name):
        return self.directory.joinpath(name)

    def test_success(self):
        runner = CliRunner()
        result = runner.invoke(cli, [
            f"--config={self.get_config('success.yml')}",
            self.executable, '-v'])

        print(f"Output: {result}")
        print(f"Executed: {self.executable}")
        print(f"Config: {self.get_config('success.yml')}")

        self.assertEqual(result.exit_code, 0)
        self.assertIn('Ran 4 tests in', result.output)
        self.assertIn('ok.', result.output)

    def test_failure(self):
        runner = CliRunner()
        result = runner.invoke(cli, [
            f"--config={self.get_config('fail.yml')}",
            self.executable, '-v'])

        print(result)
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Invalid exit status', result.output)
        self.assertIn('Invalid value on stdout', result.output)
        self.assertIn('fail.', result.output)

    def test_inverse(self):
        runner = CliRunner()
        result = runner.invoke(cli, [
            f"--config={self.get_config('inverse.yml')}",
            self.executable, '-v'])

        print(result, result.output)

        self.assertEqual(result.exit_code, 0)
        self.assertIn('ok.', result.output)

    def test_ignorespaces(self):
        runner = CliRunner()
        result = runner.invoke(cli, [
            f"--config={self.get_config('ignorespaces.yml')}", '-vv'])

        print(result, result.output)

        self.assertEqual(result.exit_code, 0)
        self.assertIn('ok.', result.output)
